import { Component, OnInit, OnDestroy, ChangeDetectorRef, Directive, ViewChild } from '@angular/core';
import { Sign } from 'src/app/shared/models/sign.model';
import { Acquisition } from 'src/app/shared/models/acquisition.model';
import { AcquisitionsStoreFacade } from 'src/app/shared/state/acquisition/acquisition.store-facade';
import { AcquisitionsService } from 'src/app/shared/services/acquisitions.service';
import { Command } from 'src/app/shared/models/command.model';
import { MenuItem } from 'src/app/shared/models/menuItem.model';
import { CommandsStoreFacade } from 'src/app/shared/state/command/command.store-facade';
import { EventsStoreFacade } from 'src/app/shared/state/event/event.store-facade';
import { NGXLogger } from 'ngx-logger';
import { MatMenuTrigger, MenuPositionX, MenuPositionY } from '@angular/material/menu';
import { Subscription } from 'rxjs';
import { AppSettings } from 'src/app/shared/app.settings';

@Directive()
export class AbstractSymbolComponent implements OnInit, OnDestroy {

  @ViewChild(MatMenuTrigger)
  contextMenu: MatMenuTrigger;

  @Component({ template: '' })
  protected _rtAcqs: Acquisition[] = [];
  protected _rtCmds: Command[] = [];

  protected _symbolData: Sign;


  public signal: string;
  public currentPicture: string;
  public isValid: boolean = false;
  public currentValue: boolean | number;

  public cmdMenuItems: MenuItem[] = [];
  public contextMenuPosition = { x: '0px', y: '0px' };
  private _sub2Acqs: Subscription;
  private _sub2Cmds: Subscription;
  private _acqTimer: any = null;
  public isAcqAlive: boolean = false;

  constructor(private acqsStoreFacade: AcquisitionsStoreFacade,
    private cmdsStoreFacade: CommandsStoreFacade,
    private eventsStoreFacade: EventsStoreFacade,
    private acqSvc: AcquisitionsService,
    private cdRef: ChangeDetectorRef, private logger: NGXLogger) { }

  ngOnInit(): void {

    this.subscribeToAcqs();
    this.subscribeToCmds();
    this.subscribeToSignalWD();
  }
  ngOnDestroy() {
    this._sub2Acqs.unsubscribe();
    if (this._sub2Cmds) {
      this._sub2Cmds.unsubscribe();
    }
    clearTimeout(this._acqTimer);

  }
  subscribeToAcqs() {
    this.logger.debug('AbstractSymbolComponent: subscribeToAcqs');

    this._symbolData.acqs.map(
      acqName => {
        this._sub2Acqs = this.acqsStoreFacade.getAcquisitionsByName(acqName).subscribe(

          (updatedAcq) => {
            if (updatedAcq) {
              this.logger.debug("updatedAcq ", updatedAcq);
              this._rtAcqs = this._rtAcqs.filter(acq => acq.signal !== updatedAcq.signal);
              this._rtAcqs.push(updatedAcq);
              this.resolveConditions();
            }
          });
      });
  }
  subscribeToSignalWD() {
    this.acqSvc.watchdog.subscribe(
      (wd) => {

        if (this.isAcqAlive != wd) {
          // recovering from a communication loss
          this.isAcqAlive = wd;
          this._rtCmds = this._rtCmds.map((item) => {
            let updatedCmd = { ...item };
            updatedCmd.isValid = true;
            return updatedCmd;
          });
          this.resolveConditions();
        }

        if (this._acqTimer) {
          clearTimeout(this._acqTimer);
        };

        this._acqTimer = setTimeout(() => {
          this.isAcqAlive = false;
          this.logger.debug("invalidating acqs and cmds");
          this._rtAcqs = this._rtAcqs.map((item) => {
            let updatedAcq = { ...item };
            updatedAcq.isValid = false;
            return updatedAcq;
          });
          this._rtCmds = this._rtCmds.map((item) => {
            let updatedCmd = { ...item };
            updatedCmd.isValid = false;
            return updatedCmd;
          });

          this.resolveConditions();

        }, AppSettings.WD_TIMEOUT);
      });
  }


  subscribeToCmds() {


    // subscribe to commands

    this._symbolData.cmds.map(
      cmdFromSymbol => {
        this.logger.debug('AbstractSymbolComponent: subscribeToCmd ', cmdFromSymbol.signal);
        this._sub2Cmds = this.cmdsStoreFacade.getCommandsByName(cmdFromSymbol.signal).subscribe(

          (updatedCmd) => {
            if (updatedCmd) {
              this.logger.debug("updatedCmd ", updatedCmd);
              this._rtCmds = this._rtCmds.filter(cmd => cmd.signal !== updatedCmd.signal);
              this._rtCmds.push(updatedCmd);
              this.resolveConditions();
              this.resolveCmdMenu();
            }
          });
      });
  }

  updateStaticData(symbolData: Sign) {
    this.logger.debug('AbstractSymbolComponent: updateStaticData');
    this._symbolData = symbolData;
    this.logger.debug(this._symbolData);
    this.initializeSymbols();
    this.initializeRealTimeVariables();
    this.cdRef.detectChanges();
  }

  initializeSymbols() {
    this.logger.debug('AbstractSymbolComponent: initializeSymbols');
    // initialize symbol from condition values
    // the default picture shall have initial condition value = true in its definition
    for (const condition of this._symbolData.conditions) {
      if (condition.value) {
        this.currentPicture = this._symbolData.pathToImages + condition.picture;
        this.currentValue = condition.value;
        break;
      }
    }
    this.isValid = false;
    this.signal = this._symbolData.itemName;


  }

  initializeRealTimeVariables() {
    this.logger.trace('AbstractSymbolComponent: initializeRealTimeVariables');
    // initialize real-time acquisitions array
    this._symbolData.acqs.map(acqName => {
      this._rtAcqs.push({
        id: 0,
        signal: acqName,
        value: false,
        isValid: false
      });
    });
    this._symbolData.cmds.map(cmd => {
      this._rtCmds.push({
        id: 0,
        signal: cmd.signal,
        type: cmd.type,
        value: false,
        isValid: false
      });
      if (cmd.type === "bool") {
        this.cmdMenuItems.push({
          title: cmd.signal,
          enabled: false
        });
      } else if (cmd.type === "number") {
        this.cmdMenuItems.push({
          title: cmd.signal,
          enabled: true
        });
      }
    });
    this.logger.debug('Cmds :', this._rtCmds);
    this.logger.debug('MenuItems :', this.cmdMenuItems);
  }
  resolveConditions() {
    this.logger.trace('AbstractSymbolComponent: resolveConditions');

    // then we dynamically create the local variables allowing to interpret our expressions
    // we leverage the loop to update symbol validity
    this.isValid = true;


    for (let condition of this._symbolData.conditions) {

      let condition2draw = { ...condition };
      this.logger.debug('AbstractSymbolComponent: condition:', condition2draw);

      // first of all we replace the acquisitions in the expression by its values
      for (const acq of this._rtAcqs) {

        const re = new RegExp(acq.signal, "g");
        condition2draw.expression = condition2draw.expression.replace(re, acq.value.toString());
        this.isValid = this.isValid && acq.isValid;
      }
      this.logger.debug('AbstractSymbolComponent: condition after acqs update : ' + condition2draw.expression);
      this.logger.debug('AbstractSymbolComponent: validity after acqs update : ' + this.isValid);

      // first of all we replace the acquisitions in the expression by its values
      for (const cmd of this._rtCmds) {

        const re = new RegExp(cmd.signal, "g");
        condition2draw.expression = condition2draw.expression.replace(re, cmd.value.toString());
        this.isValid = this.isValid && cmd.isValid;
      }
      this.logger.debug('AbstractSymbolComponent: condition after cmds update : ' + condition2draw.expression);
      this.logger.debug('AbbstractSymbolComponent: validity after cmds update : ' + this.isValid);


      condition2draw.value = eval(condition2draw.expression);

      if (condition2draw.value) {
        this.currentValue = condition2draw.value;
        this.currentPicture = this._symbolData.pathToImages + condition2draw.picture;
        break;
      }

    }
    this.cdRef.detectChanges();
  }
  resolveCmdMenu() {
    this.logger.trace('AbstractSymbolComponent: resolveCmdMenu');
    this.cmdMenuItems = [];

    // we disable the menu items for these commands that are enabled
    this._rtCmds.map((cmd) => {
      if (cmd.type === "bool") {
        this.cmdMenuItems.push({
          title: cmd.signal,
          enabled: !cmd.value
        });
      } else if (cmd.type === "number") {
        this.cmdMenuItems.push({
          title: cmd.signal,
          enabled: true
        });
      }

    });
    this.logger.debug("resolveCmdMenu: ", this.cmdMenuItems);
  }

  onContextMenu(event: MouseEvent) {
    event.preventDefault();
    this.contextMenuPosition.x = event.clientX + 'px';
    this.contextMenuPosition.y = event.clientY + 'px';
    this.logger.debug('Menu position ' + this.contextMenuPosition.x + ', ' + this.contextMenuPosition.y);
    this.contextMenu.menu.focusFirstItem('mouse');
    this.contextMenu.openMenu();
  }
  onContextMenuAction(item: MenuItem) {

    this._rtCmds.map(
      (cmd) => {
        let updatedCmd = { ...cmd };

        if (cmd.signal == item.title) {
          updatedCmd.value = true;
          this.cmdsStoreFacade.updateCommand(updatedCmd);
        } else {
          updatedCmd.value = false;
          this.cmdsStoreFacade.updateCommand(updatedCmd);
        }
        let event = {
          id: -1,
          type: 'COMMAND',
          timestamp: Date.now(),
          source: 'SymbolComponent',
          description: "Command " + updatedCmd.signal + " has been sent with value " + updatedCmd.value
        }
        this.eventsStoreFacade.addEvent(event);

      }
    );
  }
  onNumberControlAction(item: MenuItem, setValue: number) {
    console.log(item.title);
    console.log(setValue);

    let cmd2Update = this._rtCmds.find(cmd => cmd.signal == item.title);

    let updatedCmd = { ...cmd2Update };
    updatedCmd.value = setValue;
    this.cmdsStoreFacade.updateCommand(updatedCmd);

    let event = {
      id: -1,
      type: 'COMMAND',
      timestamp: Date.now(),
      source: 'SymbolComponent',
      description: "Command " + updatedCmd.signal + " has been sent with value " + updatedCmd.value
    }
    this.eventsStoreFacade.addEvent(event);

  }
  stopPropagation(event) {
    event.stopPropagation();
  }
}